/**
 * 
 */
package processModel;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Frame;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Observable;
import java.util.Observer;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.AbstractUndoableEdit;

import javax.swing.ImageIcon;
import javax.swing.JOptionPane;
import javax.swing.undo.UndoableEdit;

import processView.Externaliser;
import processView.ProcessApplet;

import java.io.ByteArrayOutputStream;
import java.io.PrintWriter;
import java.lang.Character;
import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;

import processView.DialogMenu;
import processView.StateDialogMenu;
import processView.TransitionDialogMenu;
import utilities.ActionEnum;
import java.text.NumberFormat;


/**
 * Process class implements the ConcreteSubject in the Observer Design Pattern.
 * The changes of its state are observed by the Controller component of the
 * applet, the ProcessApplet class, and the View component, the ProcessDiagram 
 * class. Thus the implementation of the Observer Design Pattern is closely
 * connected to the Model-View-Controller System architecture. 
 * 
 * @author Kon Zakharov, k.zakharov@cosc.canterbury.ac.nz
 */
public class Process extends Observable {

	/**
	 * Collection of diagram elements.
	 */
	private java.util.List<Transition> transitions;
	private java.util.List<StateElement> states;

	private ActionEnum mode;

	//private Element newElement; // to be made obsolete
	public Transition newTransition;
	public StateElement newState;
	private Collection<Element> elements;

	private Point pinPoint, endPoint;

	private UndoableEdit lastOperation;

	private Frame parentFrame = null;
	private boolean showRightClickMenu = false;
	public Moveable selectedObject;
	private StateElement selectedState;
	
	private Externaliser saverAndLoader;
	private ProcessApplet owner;
	
	// maintain a count of all elements used: transition handles, states and transitions
	private static int IDS_COUNT = 1;
	private boolean dontMove=false;
	private boolean firstDrag=false;
	private Point currentLocation;
	

	// Cascading Constructors.
	public Process(ProcessApplet ownerApplet) {
		super();
		owner=ownerApplet;
		saverAndLoader = new Externaliser(owner);
		transitions = new ArrayList<Transition>();
		states = new ArrayList<StateElement>();
		elements = new ArrayList<Element>();
		selectedObject = null;
	}

	public Process(Observer obs, ProcessApplet ownerApplet) {
		new Process(ownerApplet);
		addObserver(obs);
	}

	public void lastOperation(UndoableEdit lastOp){
		this.lastOperation = lastOp;
	}

	/**
	 * setState is called by the Controller - ProcessApplet.class. 
	 * For instance, when the user clicks on a button to create a new State
	 * or a transition, this method is called with the correct mode as a parameter.
	 * Note: There are two setState methods (different signatures)
	 * @param newMode ActionEnum that describes the mode
	 * 
	 */
	public void setState(ActionEnum newMode) {
		setChanged();
		mode = newMode;
		System.out.println("in setState(ActionEnum newMode) mode: " + mode);

		//this.unselectAllElements();  why?
		switch (mode) {
			case NEW_STATE:
				newState = new StateElement();
				showRightClickMenu=true;				
				//lastOperation(new AddStateEdit(newState));
				break;
			case NEW_ISOCHORIC:
				newTransition = new IsochoricTransition("isochoric");
				showRightClickMenu=true;
				break;
			case NEW_ISOBARIC:
				newTransition = new IsobaricTransition("isobaric");
				showRightClickMenu=true;
				break;
			case NEW_ISOTHERMAL:
				newTransition = new IsothermalTransition("isothermal");
				showRightClickMenu=true;
				break;
			case NEW_ADIABATIC:
				newTransition = new AdiabaticTransition("adiabatic");
				showRightClickMenu=true;
				break;			
			case EDIT_DONE:
				pinPoint = null;
				endPoint = null;
				
				if (newState != null){	
					addState(newState);
					 if (showRightClickMenu)
						this.startRightClickMenu(parentFrame, newState,(Point)(newState.getCenter()));
				}
						
					
				
				if (newTransition != null){
					addTransition(newTransition);
					if(!this.isStateTransitionUnique()){
						owner.postErrorMessage("Transitions must be unique.");
						removeTransition(newTransition);
					}
					else if (!checkTransitionVolumeIntegrity(newTransition) || !checkTransitionPressureIntegrity(newTransition)){
						owner.postErrorMessage("Please check the states you're attempting to connect.  The position of these states in the diagram is not consistent with the volume and/or pressure associated with them. ");
						removeTransition(newTransition);
					}						
					else if (showRightClickMenu) {
						// TODO:  Remove this?  Only if there have been more problems added which actually use this dialog
						if (newTransition.isNotJoined())
						{
							newTransition.setDeltaUnknown();
							newTransition.setHeatUnknown();
							newTransition.setWorkUnknown();
						}
						else {
							this.startRightClickMenu(parentFrame, newTransition, newTransition.endPoint());
						}
					}
				}
				
				
				
				// connect the appropriate elements
				// assign the appropriate transition label
				connectStatesAndTransitions();
				assignTransitionLabel();	
				
				if (newState != null){
					for(StateHandle handle: newState.getHandles()){
						for(Transition transition: getTransitionsAt(handle.origin())){
							if (!checkTransitionVolumeIntegrity(transition) || !checkTransitionPressureIntegrity(transition)){
								owner.postErrorMessage("Please check the states you're attempting to connect.  The position of these states in the diagram is not consistent with the volume and/or pressure associated with them.");
								removeState(newState);	
								break;
							}						
						}
					}
				}	
					
				elements.removeAll(elements);
				elements.addAll(states); // states first, then transitions
				elements.addAll(transitions); // transitions first, then states
				
				//diagramToASPIREXML();
				
				mode = null;
				newState = null;
				newTransition = null;
				showRightClickMenu = false;
				
				
				break;
	
			case CLEAR:
				//lastOperation(new RemoveElement(this,elements));
				if (selectedObject != null)	selectedObject.select(false);
				selectedObject = null;
				states.clear();
				transitions.clear();
				mode = null;
				newState = null;
				newTransition = null;
				showRightClickMenu = false;
				break;
			case UNDO:
				lastOperation(null);
				break;
			case REDO:
				lastOperation(null);
				break;

			case RIGHT_CLICK_MENU:
				// just setting the mode for now. Do nothing here
				break;
			case DELETE:
				System.err.println("Selected Object " + selectedObject);
				if(selectedObject != null){
					removeSelectedTransitions();
					removeSelectedStates();
				}
	
				elements.removeAll(elements);
				elements.addAll(states); // states first, then transitions
				elements.addAll(transitions); // transitions first, then states
				
				//diagramToXML();
				//System.out.println(diagramToASPIREXML());
				
				mode = null;
				newTransition = null;
				newState = null;
				showRightClickMenu = false;
				break;
			default:
				System.err.println("Do not know what to do with " + mode);
		}
		// This call tells the registered observers it's time to ask for
		// updated state. For now observers don't do much, but it has to change.
		notifyObservers("Completed processing setState(" + newMode + ")");
	}
	
	private void addState(StateElement newState) {
		if (newState != null){
			newState.setID(getNextID());
			states.add(newState);			
			// record the effect
            lastOperation(new AddStateEdit(newState));
		}
	  }
	
	public static int getNextID()
	{
		return IDS_COUNT++;
	}
	
	public void removeState(StateElement state) {
		Iterator <StateHandle> handleIt = state.getHandles().iterator();
		
		while(handleIt.hasNext()){
			Handle handle = handleIt.next();
			if(!getTransitionsAt(handle.origin()).isEmpty())
			{	
				java.util.List<Transition> matchedTransitions = getTransitionsAt(handle.origin());	
				for(Transition transition : matchedTransitions){
					for (TransitionHandle transitHandle : transition.getHandles()){
						if(transitHandle.origin() == handle.origin()) transitHandle.disconnect();
					}
				}
			}
		}
		
		//states.get(states.indexOf(state)).prepareForDeletion();		
		states.remove(state);
	}
	
	public void addTransition(Transition newTransition) {
		if (newTransition != null){			
			newTransition.setID(getNextID());
			//IDS_COUNT = IDS_COUNT + 2; // account for the id's of the points corresponding to this transition
			transitions.add(newTransition);
			
			// record the effect
            lastOperation(new AddTransitionEdit(newTransition));
		}
	}
	
	public void removeTransition(Transition transition) {
		//transition.prepareForDeletion();
		transitions.remove(transition);
	}
	

	public void removeSelectedStates() {
		
		List <StateElement> statesToDelete = new ArrayList <StateElement> ();
		System.out.println("DELETING Tstate");
		for (StateElement tempState: states){
			if (tempState != null && tempState.selected()){
				statesToDelete.add(tempState);
			}
		} 		
		
		//		 Then delete states
		for (StateElement tempState : statesToDelete){
			//	record the effect
            lastOperation(new RemoveStateEdit(tempState));
			
			removeState(tempState);
		}
	}
	
	public void removeSelectedTransitions() {
		List <Transition> transitionsToDelete = new ArrayList <Transition> ();
		System.out.println("DELETING Transition");
		// Get the selected elements
		for (Transition tempTransition: transitions){
			if (tempTransition != null && tempTransition.selected()){
				transitionsToDelete.add(tempTransition);
			}
		}		

		// First delete the transitions. This would take care of connectivity
		for (Transition tempTransition : transitionsToDelete){
			//	record the effect
			lastOperation(new RemoveTransitionEdit(tempTransition));			
			
			removeTransition(tempTransition);
		}
	}
	
	  public class AddStateEdit extends AbstractUndoableEdit {		    
		  	private StateElement state;
		  	//private List<StateHandle> handles;
		  
		  	public AddStateEdit(StateElement state) {
			  this.state=state;
			 // this.handles = state.getHandles();
	    	}
	
		    public void undo() throws CannotUndoException {
		    	removeState(state);
		    }
		
		    public void redo() throws CannotRedoException {
		    	//state.setHandles(handles);
		    	addState(state);	    	
		    }
		    
		    public boolean canUndo() {
		         return true;
		    }
	
		    public boolean canRedo() {
		         return true;
		    }
		
		    public String getPresentationName() {
		     return "";
		    }
	  }
	  
	  public class RemoveStateEdit extends AbstractUndoableEdit {		    
		  	private StateElement state;
		  	//private List<StateHandle> handles;
		  
		  	public RemoveStateEdit(StateElement state) {
			  this.state=state;
			 // this.handles = state.getHandles();
	    	}
	
		    public void undo() throws CannotUndoException {
		    	addState(state);
		    }
		
		    public void redo() throws CannotRedoException {
		    	//state.setHandles(handles);
		    	removeState(state);	    	
		    }
		    
		    public boolean canUndo() {
		         return true;
		    }
	
		    public boolean canRedo() {
		         return true;
		    }
		
		    public String getPresentationName() {
		     return "";
		    }
	  }
	  
	  public class AddTransitionEdit extends AbstractUndoableEdit {		    
		  private Transition transition;
		  	private List<TransitionHandle> handles;
		  
		  	public AddTransitionEdit(Transition transition) {
			  this.transition=transition;
			  this.handles = transition.getHandles();
	    	}
	
		    public void undo() throws CannotUndoException {
		    	removeTransition(transition);
		    }
		
		    public void redo() throws CannotRedoException {
		    	transition.setHandles(handles);
		    	addTransition(transition);	    	
		    }
		    
		    public boolean canUndo() {
		         return true;
		    }
	
		    public boolean canRedo() {
		         return true;
		    }
		
		    public String getPresentationName() {
		      return "";
		    }
	  }
	  
	  public class RemoveTransitionEdit extends AbstractUndoableEdit {		    
		  private Transition transition;
		  	private List<TransitionHandle> handles;
		  
		  	public RemoveTransitionEdit(Transition transition) {
			  this.transition=transition;
			  this.handles = transition.getHandles();
	    	}
	
		    public void undo() throws CannotUndoException {
		    	transition.setHandles(handles);
		    	addTransition(transition);
		    }
		
		    public void redo() throws CannotRedoException {
		    	removeTransition(transition);	    	
		    }
		    
		    public boolean canUndo() {
		         return true;
		    }
	
		    public boolean canRedo() {
		         return true;
		    }
		
		    public String getPresentationName() {
		      return "";
		    }
	  }
	

	/**
	 * setState is called by the Controller - ProcessApplet class.
	 * Given a particular point where the event occurred and a type of event (MouseEvent),
	 * this method sets a particular state for the process. 
	 * @param p Point where the event occured
	 * @param eventType the type of event given by MouseEvent
	 */
	public void setState(Point p, int eventType) {
		setChanged();
		boolean connected=false;
		// Here we need to decide what to do with the new point.
		if (pinPoint == null) {
			// This is what we do if the this is a pinPoint.
			pinPoint = new Point(p);
			if (mode != null){
				switch (mode) {	
				case NEW_STATE:
					newState.setCenter(pinPoint);
					break;
				case NEW_ISOCHORIC:
				case NEW_ISOBARIC:
				case NEW_ISOTHERMAL:
				case NEW_ADIABATIC:
					newTransition.setOrigin(pinPoint);
					break;
				case RIGHT_CLICK_MENU:
					// Do right click menu stuff here!!
					break;
				default: break;
				}
				
			}else
				selectAndHighlightElement(pinPoint);
		} else {
			// This is what we do if this is an endPoint.
			System.out.println("HERE - endpoint");
			endPoint = new Point(p);
			if (mode != null){
				switch (mode) {
				case NEW_STATE:
					newState.setCenter(pinPoint);
					break;
				case NEW_ISOCHORIC:
					if(pinPoint.x!=p.x || pinPoint.y!=p.y)
						newTransition.setEndPoint(endPoint);
					break;
				case NEW_ISOBARIC:
				case NEW_ISOTHERMAL:
				case NEW_ADIABATIC:
					if(pinPoint.x!=p.x || pinPoint.y!=p.y)
						newTransition.setEndPoint(endPoint);
					break;
				}
				
				if(newTransition!= null && newTransition.type.equals("isochoric") && eventType == MouseEvent.MOUSE_RELEASED){					
					if(newTransition!=null && newTransition.endPoint() != null && newTransition.beginPoint() != null){							
						if(newTransition.endPoint().x > newTransition.beginPoint().x + 20 ||
								newTransition.endPoint().x < newTransition.beginPoint().x  - 20){
							owner.postErrorMessage("An Isochoric transition is a constant volume transition.  Therefore, the transition should be parallel to the y-axis.");
							newTransition=null;
						}
					}
				}
				else if(newTransition!= null && newTransition.type.equals("isobaric") && eventType == MouseEvent.MOUSE_RELEASED){					
					if(newTransition!=null && newTransition.endPoint() != null && newTransition.beginPoint() != null){							
						if(newTransition.endPoint().y > newTransition.beginPoint().y + 20 ||
								newTransition.endPoint().y < newTransition.beginPoint().y  - 20){
							owner.postErrorMessage("An Isobaric transition is a constant pressure transition.  Therefore, the transition should be parallel to the x-axis.");
							newTransition=null;
						}
					}
				}
				else if(newTransition != null && (newTransition.type.equals("isothermal") || newTransition.type.equals("adiabatic")) && eventType == MouseEvent.MOUSE_RELEASED)
				{
					if (newTransition.isInvalidExpansionOrCompression())
					{
						owner.postErrorMessage("An isothermal or adiabatic transition is not a valid expansion or compression.</p>" +
								"<p>If pressure is increasing, volume must be decreasing and vice versa.");
						newTransition = null;
					}
				}
			}
			else if (eventType == MouseEvent.MOUSE_DRAGGED){
				if(firstDrag==false){
					currentLocation = selectedState.getCircleBounds().getLocation();	
					for(StateHandle handle: selectedState.getHandles())
					{	//mark whether the handle is connected or not before it is dragged
						if(handle.isConnected()) handle.setWasConnected();
						System.out.println("wasConnected:" + handle.wasConnected());
					}
					firstDrag=true;
				}
					
				this.changeObjectLocation(p);
				}
			else if(eventType == MouseEvent.MOUSE_RELEASED){
				connectStatesAndTransitions();
				String displayMessage= "Please check the states you're attempting to move.  The position of these states in the diagram is not consistent with the volume and/or pressure associated with them.";
				
				StateElement state = getStateElementAt(p);
				if(state!=null){
					for(StateHandle handle: state.getHandles()){
						for(Transition transition: getTransitionsAt(handle.origin())){
							if (!checkTransitionVolumeIntegrity(transition) || !checkTransitionPressureIntegrity(transition)){
								System.out.println("wasConnected: isConnected" + handle.wasConnected() + handle.isConnected());
								if(!handle.wasConnected() && handle.isConnected()){
									if(transition.getHandles().get(0).getPartner() == handle) transition.getHandles().get(0).disconnect();
									else transition.getHandles().get(1).disconnect();
									//need to redisconnect the statehandle since its no longer connected
									handle.setConnected(false);
								}								
								dontMove=true;
								break;
							}	
							System.out.println("TRANSITION TYPE: " + transition.type);
							if(transition.type.equals("isochoric")){	
								if(transition!=null && transition.endPoint() != null && transition.beginPoint() != null){							
									if(transition.endPoint().x > transition.beginPoint().x + 20 ||
											transition.endPoint().x < transition.beginPoint().x  - 20){
										dontMove=true;
										displayMessage="An Isochoric transition is a constant volume transition.  The transition is therefore perpendicular to the x-axis.";
										break;
										}
									}
								}
							else if(transition.type.equals("isobaric")){	
								if(transition!=null && transition.endPoint() != null && transition.beginPoint() != null){							
									if(transition.endPoint().y > transition.beginPoint().y + 20 ||
											transition.endPoint().y < transition.beginPoint().y  - 20){
										dontMove=true;
										displayMessage="An Isobaric transition is a constant pressure transition.  The transition is therefore perpendicular to the y-axis.";
										break;
									}
								}
							}
							else if(transition.type.equals("isothermal") || transition.type.equals("adiabatic"))
							{
								if (transition.isInvalidExpansionOrCompression())
								{
									displayMessage = "An isothermal or adiabatic transition is not a valid expansion or compression.</p>" +
													 "<p>If pressure is increasing, volume must be decreasing and vice versa.";
									dontMove = true;
									break;
								}
							}
						}
					}
				}
				
				System.out.println("Mouse Released: dont move" + dontMove + currentLocation);
				if(dontMove && currentLocation!=null){					
					this.changeObjectLocation(currentLocation);
					owner.postErrorMessage(displayMessage);
				}
				dontMove=false;
				firstDrag=false;
				currentLocation=null;
			}
		}	

		// This call tells the registred observers it's time to ask for
		// updated state. For now observers don't do much, but it has to change.

		notifyObservers("Completed processing setState(" + p + ")");
	}
	
	public boolean checkTransitionVolumeIntegrity (Transition transition)
	{
		StateElement beginState = getStateElementAt(transition.beginPoint());
		StateElement endState = getStateElementAt(transition.endPoint());
		
		NumberFormat nf = NumberFormat.getNumberInstance();

		if(endState==null || beginState==null) return true;	
		
		if (endState.volume()==null || (endState.volume() + " ").equals(" ") || beginState.volume()==null || (beginState.volume() + " ").equals(" ")) return true;
		
		try{
			if(nf.parse(endState.volume()).floatValue() == nf.parse(beginState.volume()).floatValue()
					&& (transition.endPoint().x < transition.beginPoint().x + 20 && 
							transition.endPoint().x > transition.beginPoint().x  - 20)) 
				return true;
			else if(transition.endPoint().x < transition.beginPoint().x
					&& nf.parse(endState.volume()).floatValue() < nf.parse(beginState.volume()).floatValue())
				return true;				 
			else if (transition.endPoint().x > transition.beginPoint().x
					&& nf.parse(endState.volume()).floatValue() > nf.parse(beginState.volume()).floatValue())
				return true;
			
			else return false;
		} catch (Exception e){
			return false;
		}
	}
	
	public boolean checkTransitionPressureIntegrity (Transition transition)
	{
		StateElement beginState = getStateElementAt(transition.beginPoint());
		StateElement endState = getStateElementAt(transition.endPoint());
		
		NumberFormat nf = NumberFormat.getNumberInstance();

		if(endState==null || beginState==null) return true;		
		if (endState.pressure()==null || (endState.pressure()+ " ").equals(" ") || beginState.pressure()==null || (beginState.pressure() + " ").equals(" ")) return true;

		try{
			if(nf.parse(endState.pressure()).floatValue() == nf.parse(beginState.pressure()).floatValue()
					&& (transition.endPoint().y > transition.beginPoint().y - 20 &&
							transition.endPoint().y < transition.beginPoint().y + 20))
				return true;
			else if(transition.endPoint().y < transition.beginPoint().y
					&& nf.parse(endState.pressure()).floatValue() > nf.parse(beginState.pressure()).floatValue())
				return true;				 
			else if (transition.endPoint().y > transition.beginPoint().y
					&& nf.parse(endState.pressure()).floatValue() < nf.parse(beginState.pressure()).floatValue())
				return true;
			else return false;
		} catch (Exception e){
			return false;
		}
	}


	/**
	 * Given a particular point, this method checks to see if that point falls inside
	 * any element. If so, it asks the element to select itself (TellDontAsk) which
	 * in turn displays the selected element appropriately.
	 * and adding it to the selectedObject.
	 * @param p
	 * @param selectTheElement
	 */
	public void selectAndHighlightElement(Point p){

		try{ 
			selectedObject.select(false);
		} catch (Exception e){}
		selectedObject = null;		

		Collection<Element> elements = new ArrayList<Element>();

		elements.addAll(states); // states first, then transitions
		elements.addAll(transitions); // transitions first, then states

		Handle selectedHandle = null;
		Element selectedElement = null;
		try{
			parentFrame.getComponentAt(p).setCursor(Cursor.getDefaultCursor());
		} catch (Exception e){}

		// go through the TransitionHandles first
		for (Element tempElement : elements){			
			selectedElement = tempElement.pointIsInElement(p);
			if (selectedElement != null){
				tempElement.select(true);
				//parentFrame.getComponentAt(p).setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
				break;
			} else {
				selectedHandle = tempElement.pointIsInHandle(p);
				if (selectedHandle != null){
					selectedElement = selectedHandle.getParent();
					selectedElement.select(true);
					//parentFrame.getComponentAt(p).setCursor(selectedHandle.getCursorType());
					break;
				}
			}
		}
		// Assuming that all elements can be moved!
		if (selectedElement != null) {
			selectedObject = (Moveable)selectedElement;
			try{
				selectedState = (StateElement)selectedElement;}
			catch (Exception e) {
			}
		}

	}

	/**
	 * changeObjectLocation asks the selectedObject (if one is selected)
	 * to move itself to a new location p
	 * @param p The new location Point
	 */
	private void changeObjectLocation(Point p){
		if (selectedObject != null){
			selectedObject.move(p);
		}
	}

	/**
	 * assignNextLabel assigns the element with a default label that is unique.
	 * @return String - a next valid label
	 */
	public String assignNextLabel(){
		// first find the total number of states and transitions
		// assigned labels are integer values. Students can assign strings
		System.out.println("About to find NEXTLABEL");
		Integer nextLabel = 0;
		Collection <Element> allElements = new ArrayList <Element>();
		try{
			allElements.addAll(states);
			nextLabel = allElements.size();
		}catch (Exception e){}

		boolean foundNextLabel = false;

		// Keep incrementing the label until an unused one is found
		while (!foundNextLabel){
			if (isLabelUsed(nextLabel.toString()))
				nextLabel++;
			else {
				foundNextLabel = true;
			}
		}
		System.out.println("FOUND NEXT LABEL *** " + nextLabel);
		return nextLabel.toString();
	}
	
	public void assignTransitionLabel(){
		for (Transition currentTransition:transitions){
			TransitionHandle tHandle1 = null;
			TransitionHandle tHandle2 = null;

			try{
				tHandle1 = currentTransition.getHandles().get(0);
				tHandle2 = currentTransition.getHandles().get(1);
				
				if(tHandle1.isConnected() & tHandle2.isConnected())
					currentTransition.label(getStateElementAt(tHandle1.getPartner().origin()).label()+ " - " + getStateElementAt(tHandle2.getPartner().origin()).label());
				else if (tHandle1.isConnected()){
					currentTransition.label(getStateElementAt(tHandle1.getPartner().origin()).label());
				}
				else if (tHandle2.isConnected()){
					currentTransition.label(getStateElementAt(tHandle2.getPartner().origin()).label());
				}

			}catch (Exception e ){}
		}
	}
	
	public boolean isStateTransitionUnique(){
		for(Transition testTransition:transitions){
			for (Transition currentTransition:transitions){	
				try{
					if(testTransition != currentTransition){						
						if(getStateElementAt(testTransition.getHandles().get(0).origin()) == getStateElementAt(currentTransition.getHandles().get(0).origin()) 
								&& getStateElementAt(testTransition.getHandles().get(1).origin()) == getStateElementAt(currentTransition.getHandles().get(1).origin())
								&& (testTransition.getHandles().get(0).getPartner()!= null)			
								&& (testTransition.getHandles().get(1).getPartner()!= null))
							
							return false;
						}
					if(getStateElementAt(testTransition.getHandles().get(0).origin()) == getStateElementAt(currentTransition.getHandles().get(1).origin()) 
							&& getStateElementAt(testTransition.getHandles().get(1).origin()) == getStateElementAt(currentTransition.getHandles().get(0).origin())
							&& (testTransition.getHandles().get(0).getPartner()!= null)			
							&& (testTransition.getHandles().get(1).getPartner()!= null))
						
						return false;
					}
				catch (Exception e ){}}
			}
		return true;
		}

	/*
	 * Checks if the String checkLabel is currently an element label
	 * @param checkLabel a String label that we are checking
	 */
	public boolean isLabelUsed(String checkLabel){
		Collection <Element> allElements = new ArrayList <Element>();
		try{
			allElements.addAll(states);
			allElements.addAll(transitions);
		}catch (Exception e){}

		for (Element e : allElements){
			if (e.label() != null)
				if (e.label().equalsIgnoreCase(checkLabel)){
					return true;
				}
		}
		return false;
	}
	/*
	 * Checks if the String checkLabel is a valid label i.e. a single digit
	 * @param checkLabel a String label that we are checking
	 */
	public boolean isLabelValid(String checkLabel){
		if(checkLabel.length()!=1) return false;
		if(Character.isDigit(checkLabel.charAt(0)))
				return true;
		else return false;
	}

/**
 * matchPointToStateElement takes a particular point as an input
 * and checks if the point falls in a StateElement. If so, it returns 
 * the stateElement, otherwise it returns null.
 * @param p The Point to check
 * @return StateElement if the point falls in a StateElement, or null
 */
	private StateElement getStateElementAt (Point p){
		StateElement sElement = null;
		for (StateElement element: states){
			sElement = (StateElement) element.pointIsInElement(p);
			if (sElement != null)
				return sElement;
		}
		return sElement;
	}

	/**
	 * matchPointToTransitionElement takes a particular point as an input
	 * and checks if the point falls in a TransitionElement. If so, it returns 
	 * the TransitionElement, otherwise it returns null.
	 * @param p The Point to check
	 * @return TransitionElement if the point falls in a TransitionElement, or null
	 */
	private Transition getTransitionAt (Point p){
		Transition tElement = null;

		for (Transition element: transitions){
			tElement = (Transition) element.pointIsInElement(p);
			if (tElement != null)
				return tElement;
		}
		return tElement;
	}

	/**
	 * matchPointToTransitionElements takes a particular point as an input
	 * and checks if the point falls in one or more TransitionElement. If so, 
	 * it returns the TransitionElement (s), otherwise it returns null.
	 * @param p The Point to check
	 * @return TransitionElement (s) if the point falls in a TransitionElement, or null
	 */
	
	public java.util.List<Transition> getTransitionsAt (Point p){
		
		java.util.List<Transition> matchedTransitions = new ArrayList<Transition>();

		for (Transition element: transitions){
			if((Transition)element.pointIsInElement(p) != null)
				matchedTransitions.add((Transition)element.pointIsInElement(p));
		}
		return matchedTransitions;
	}
	
	public void parentFrame(Frame f){this.parentFrame = f;}
	public Frame parentFrame(){return this.parentFrame;}

	/**
	 * startRightClickMenu creates and displays the dialog menu for a Transition 
	 * e.g. when a new Element (Transition) is created. The user can then 
	 * enter/modify values for the selected Element
	 * @param parent
	 * @param transition
	 * @param pointOfClick
	 */
	public void startRightClickMenu(Frame parent, Transition transition, Point pointOfClick){
		
		if (transition == null){
			return;
		}
		DialogMenu d = null;
		if (transition != null && this.getTransitionAt(pointOfClick)!= null){
			String title = "Set " + (transition.label()!= null?transition.label():"") + " Values";
			d = new TransitionDialogMenu(parent, title, transition, this); // using tell-dont-ask 
		}
		else return;
		if (d != null){
			d.pack(); 
			Point appLocation = owner.getLocationOnScreen();
			d.setLocation(appLocation.x+owner.getWidth()/2-d.getWidth()/2, appLocation.y);
			d.setVisible(true);
		} 
	}

	/**
	 * startRightClickMenu creates and displays the dialog menu for a StateElement
	 * e.g. when a new Element (StateElement) is created. The user can then 
	 * enter/modify values for the selected Element
	 * @param parent
	 * @param transition
	 * @param pointOfClick
	 */
	public void startRightClickMenu (Frame parent, StateElement state, Point pointOfClick){
		if (state == null)return;

		DialogMenu d = null;
		if (state != null && this.getStateElementAt(pointOfClick)!=null){
			String title = "Set " + (state.label()!= null?state.label():"") + " Values";
			d = new StateDialogMenu (parent, title, state, this); // using tell-dont-ask
		}
		else return;
		if (d != null){
			d.pack(); 
			Point appLocation = owner.getLocationOnScreen();
			d.setLocation(appLocation.x+owner.getWidth()/2-d.getWidth()/2, appLocation.y);
			d.setVisible(true);
		} 
	}

	/**
	 * startRightClickMenu creates and displays the dialog menu for the element 
	 * Given a particular point, this method checks to see if the point falls within an 
	 * element. If so, it displays the correct dialog.
	 * @param parent
	 * @param pointOfClick
	 */
	public void startRightClickMenu (Frame parent, Point pointOfClick) {
		if (parent == null)return;
		DialogMenu d = null;
		StateElement state = this.getStateElementAt(pointOfClick);
		if (state != null){	
			String title = "Set State " + (state.label()!= null?state.label():"")+" Values";
			d = new StateDialogMenu (parent, title, state, this);
		}
		else {
			Transition transition = this.getTransitionAt(pointOfClick);
			if (transition != null){
				String title = "Set Transition " + (transition.label()!= null?transition.label():"") + " Values";
				d = new TransitionDialogMenu(parent, title, transition, this); 
			}
		}
		
		if (d != null){
			d.pack();
			Point appLocation = owner.getLocationOnScreen();
			d.setLocation(appLocation.x+owner.getWidth()/2-d.getWidth()/2, appLocation.y);
			d.setVisible(true);
		} 
	}

	/**
	 * unselectAllElements goes through all the Elements ad asks them
	 * to unselected themselves.
	 *
	 */
	/*private void unselectAllElements(){
		for (Moveable state : states)
			state.select(false);
		for(Moveable transition : transitions)
			transition.select(false);
	}*/

	/**
	 * connectStatesAndTransitions goes through all the elements and
	 * makes sure that any elements that overlap on the diagram are connected.
	 * This ensures proper connectivity.
	 *
	 */
	private void connectStatesAndTransitions(){
		for (Transition currentTransition:transitions){
			TransitionHandle tHandle1 = null;
			TransitionHandle tHandle2 = null;

			try{
				tHandle1 = currentTransition.getHandles().get(0);
				tHandle2 = currentTransition.getHandles().get(1);
				if (!tHandle1.isConnected()){
					Point firstPoint = tHandle1.origin();
					StateElement stateToConnectTo = getStateElementAt(firstPoint);
					if (stateToConnectTo != null){
						tHandle1.connect(stateToConnectTo.getClosestHandle(firstPoint));
					}
				}

				if (!tHandle2.isConnected()){
					Point secondPoint = tHandle2.origin();
					StateElement stateToConnectTo = getStateElementAt(secondPoint);
					if (stateToConnectTo != null){
						tHandle2.connect(stateToConnectTo.getClosestHandle(secondPoint));
					}
				}

			}catch (Exception e ){}
		}
	}

	public void paint(Graphics2D g2) {

		// first paint all the states, THEN all the transitions

		Iterator <StateElement> sIter = states.iterator();
		while (sIter.hasNext())
			sIter.next().paint(g2);

		Iterator <Transition> tIter = transitions.iterator();
		while (tIter.hasNext())
			tIter.next().paint(g2);


		// Draw the current temp element.
		g2.setColor(Color.RED);
		if (newState != null) 
			newState.paint(g2);
		else
			if (newTransition != null)
				newTransition.paint(g2);
	}

	public UndoableEdit getLastEdit() {		
		System.out.println("In process.getLastEdit, lastOperation is: " + lastOperation);
		return lastOperation; 
	}

	/*
	protected void removeElement(Element element) {
		setChanged();
		if (elements.contains(element)){
			elements.remove(element);
			notifyObservers ("completed processing removeElement " + element);
		}
		else 
			System.out.println("In process.removeElement, element " + element + " is not in elements");

	}


	protected void addElement(Element element) {
		elements.add(element);
	}

	class ClearAction extends AbstractAction {
		private static final long serialVersionUID = 2861432207980704461L;

		public void actionPerformed(ActionEvent e) {
			elements.clear();
		}
	}
	 */


	/*
	class DeleteAction extends AbstractAction {
		private static final long serialVersionUID = -3684273476805195579L;

		public void actionPerformed(ActionEvent e) {
			Element toRemove = null;
			Element current;
			Iterator<Element> iterator = elements.iterator();
			while (iterator.hasNext()) {
				current = iterator.next();
				if (((Moveable) current).selected()) {
					toRemove = current;
					break;
				}
			}
			if (toRemove != null) {
				elements.remove(toRemove);
			}
		}
	}
	 */
	
	public java.util.List<StateElement> getStates(){
		return states;
	}
	
	public java.util.List<Transition> getTransitions(){
		return transitions;
	}

	/*public void diagramToXML() {
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		getExternaliser().saveToXML(new PrintWriter(bos), owner.getUserName(), owner.getOntologyID(), elements);
		System.out.println(bos.toString());
	}*/
	
	public String diagramToASPIREXML() {
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		getExternaliser().saveDiagramToASPIREXML(new PrintWriter(bos), owner.getUserName(), owner.getOntologyID(), states, transitions);
		return bos.toString();
	}
	
	public Externaliser getExternaliser()
	{
		return saverAndLoader;
	}
	
	public List<Unknown> getUnknowns() {
		ArrayList<Unknown> retList = new ArrayList<Unknown>();
		Iterator<Transition> transitionIt = transitions.iterator();
		Iterator<StateElement> stateIt = states.iterator();
		
		while (stateIt.hasNext()) {
			retList.addAll(stateIt.next().getUnknowns());
		}
		
		while (transitionIt.hasNext()) {
			retList.addAll(transitionIt.next().getUnknowns());
		}
		
		return retList;
	}
	
	public String getUnknownsXML() {
		String xmlStr = "<unknowns>";
		Iterator it = getUnknowns().iterator();		
		while (it.hasNext()) {
			xmlStr += "<label>" + it.next() + "</label>";
		}		
		xmlStr += "</unknowns>";
		return xmlStr;
	}
	
	public static String roundNumber(String number)
	{
		return roundNumber(number, 3);
	}
	
	public static String roundNumber(String number, int sigfigs)
	{
		if (number.length() == 0) {
			return "";
		}
		try {
			MathContext mc = new MathContext(sigfigs, RoundingMode.HALF_DOWN);
			BigDecimal bigDecimal = new BigDecimal(number, mc);
			return bigDecimal.toPlainString();
		}
		catch (NumberFormatException ex) {
			System.err.println("Failed to round number due to invalid format");
			ex.printStackTrace();
			return number;
		}
	}
}
